import json


def load_networks() -> list:
    networks = []
    with open('data/networks.json', 'r') as f:
        network_map = json.load(f)

    for network in network_map.get('networks'):
        networks.append(network.get('name'))
    return networks



from web3 import Web3
from web3.exceptions import BadFunctionCallOutput
from eth_utils import to_checksum_address


def validate_addr(addr):
    try:
        to_checksum_address(addr)
    except ValueError:
        return False
    return True


def valid_contract(w3: Web3, addr):
    try:
        assert w3.eth.getCode(addr).hex() != '0x'
    except AssertionError:
        return False
    return True


def valid_token(token):
    try:
        token.functions.balanceOf(token.address).call()
    except BadFunctionCallOutput:
        return False
    else:
        return True


#!/usr/bin/env python3
import argparse
import os
import uuid

import dotenv
import web3
from web3.middleware import geth_poa_middleware

import lib.style
from lib import abi_lib
from lib.wallet_manager import WalletManager


class SecureWeb3:
    def __init__(self, wallet_file=None, network='ethereum'):
        dotenv.load_dotenv()
        self.endpoint = None
        self.token_abi = None
        self.account = None
        self._print = lib.style.PrettyText(0)
        self.wallet_file = wallet_file
        self.network = network
        self.w3 = self.setup_w3()

    def setup_w3(self):
        """
        Configure the web3 rpc endpoint object. Place any chain specific variables here.
        :return: web3.Web3()
        """
        w3_endpoint = os.environ.get(f'{self.network}_http_endpoint')
        self.w3 = web3.Web3(web3.HTTPProvider(w3_endpoint))
        if self.w3.isConnected:
            self._print.good(f"Connected to chain: {self.w3.eth.chain_id}")
        else:
            self._print.error(f'Web3 could connect to remote endpoint: {w3_endpoint}')
        if self.network == 'ethereum':
            self.endpoint = 'https://api.0x.org/'
            self.token_abi = lib.abi_lib.EIP20_ABI
        elif self.network == 'polygon':
            self.endpoint = 'https://polygon.api.0x.org/'
            self.token_abi = lib.abi_lib.EIP20_ABI
            self.w3.middleware_onion.inject(geth_poa_middleware, layer=0)
        elif self.network == 'bsc':
            self.token_abi = lib.abi_lib.BEP_ABI
            # self.w3.middleware_onion.inject(geth_poa_middleware, layer=0)
            self.endpoint = 'https://bsc.api.0x.org/'
            self._print.warning('Connected to BSC, which has not been tested very well yet.')
        elif self.network == 'aurora':
            self.token_abi = lib.abi_lib.EIP20_ABI

        self._print.good(f'Web3 connected to chain: {self.w3.eth.chain_id}')
        return self.w3

    def load_wallet(self):
        """
        Load the wallet manager and decrypt the wallet.
        :return:
        """
        if not self.wallet_file:
            self.wallet_file = os.environ.get('default_wallet_location')
        w = WalletManager(self.wallet_file)
        # w.setup_wizard()
        self.account = web3.Account.from_key(w.decrypt_load_wallet())
        if self.account:
            # overwrite in memory
            w = uuid.uuid4().hex
            del w
            return True
        return False

    def configure_wallet(self):
        """
        Launch the setup wizards and store the encrypted wallet.
        :return:
        """
        w = WalletManager(self.wallet_file)
        w.setup_wizard()
        w = uuid.uuid4().hex
        del w

    @property
    def web3(self):
        """
        Place logic for using different endpoints in different scenarios here.
        :return:
        """
        return self.w3

    def switch_network(self, network_name, poa=False):
        """
        Switch the chain.
        :param network_name:
        :param poa:
        :return:
        """
        endpoint = os.environ.get(f'infura_{network_name}_endpoint')
        if not endpoint:
            self._print.error('Could not find network, is it configured?')
            return
        w3 = web3.Web3(web3.HTTPProvider(endpoint))
        if network_name == 'polygon':
            poa = True
        if poa:
            w3.middleware_onion.inject(geth_poa_middleware)
        if w3.isConnected():
            self._print.good(f'Successfully switched to {network_name}')
        else:
            self._print.error(f'Web3 could not connect to endpoint at {endpoint}')


if __name__ == '__main__':
    args = argparse.ArgumentParser('Secure Web3 Cli')
    args.add_argument('-i', '--init', dest='init_wallet', type=str, nargs=1,
                      default=None, help='Initialize this new wallet')
    args.add_argument('-o', '--open', dest='open_wallet', type=str, default=None,
                      nargs='?', help='Unlock a wallet by specifying'
                                      'an environment variable name, '
                                      'use default wallet if not specified.')
    args.add_argument('-l', '--lock', dest='lock_wallet', action='store_true',
                      help='Lock all open wallets.')
    args.add_argument('-n', '--network', type=str, default='ethereum',
                      choices=['ethereum', 'polygon', 'bsc', 'aurora'])
    args = args.parse_args()
    dotenv.load_dotenv()

    if args.init_wallet:
        print(f'[*] Configuring new wallet "{args.init_wallet[0]}" ... ')
        manager = WalletManager(args.init_wallet[0])
        manager.setup_wizard()
        exit(0)
    if not args.open_wallet:
        wallet_file = os.environ.get('default_wallet_location')
    else:
        wallet_file = args.open_wallet
    print(f'[*] Opening wallet "{wallet_file}" ... ')
    manager = WalletManager(wallet_file)
    priv_key = manager.decrypt_load_wallet()
    # Do something with private key ...
    # print(priv_key)
    print('[+] Successfully unlocked wallet!')



from eth_utils import to_checksum_address
from lib import w3_validation
from lib import style
s = style.PrettyText(1)


def get_dest_addr():
    while True:
        destination = input("Destination Address >> ")
        if w3_validation.validate_addr(destination):
            return to_checksum_address(destination)
        s.error(f'Invalid address: "{destination}"')


def confirmation(prompt='Confirm? y/n >> '):
    try:
        confirm_yn = input(prompt)
    except KeyboardInterrupt:
        s.warning('User canceled action')
    else:
        if confirm_yn.lower() in ['yes', 'y']:
            return True
        return False




import json

EIP20_ABI = json.loads('[{"constant":true,"inputs":[],"name":"name","outputs":[{"name":"","type":"string"}],'
                       '"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{'
                       '"name":"_spender","type":"address"},{"name":"_value","type":"uint256"}],"name":"approve",'
                       '"outputs":[{"name":"","type":"bool"}],"payable":false,"stateMutability":"nonpayable",'
                       '"type":"function"},{"constant":true,"inputs":[],"name":"totalSupply","outputs":[{"name":"",'
                       '"type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},'
                       '{"constant":false,"inputs":[{"name":"_from","type":"address"},{"name":"_to",'
                       '"type":"address"},{"name":"_value","type":"uint256"}],"name":"transferFrom","outputs":[{'
                       '"name":"","type":"bool"}],"payable":false,"stateMutability":"nonpayable","type":"function"},'
                       '{"constant":true,"inputs":[],"name":"decimals","outputs":[{"name":"","type":"uint8"}],'
                       '"payable":false,"stateMutability":"view","type":"function"},{"constant":true,"inputs":[{'
                       '"name":"_owner","type":"address"}],"name":"balanceOf","outputs":[{"name":"",'
                       '"type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},'
                       '{"constant":true,"inputs":[],"name":"symbol","outputs":[{"name":"","type":"string"}],'
                       '"payable":false,"stateMutability":"view","type":"function"},{"constant":false,"inputs":[{'
                       '"name":"_to","type":"address"},{"name":"_value","type":"uint256"}],"name":"transfer",'
                       '"outputs":[{"name":"","type":"bool"}],"payable":false,"stateMutability":"nonpayable",'
                       '"type":"function"},{"constant":true,"inputs":[{"name":"_owner","type":"address"},'
                       '{"name":"_spender","type":"address"}],"name":"allowance","outputs":[{"name":"",'
                       '"type":"uint256"}],"payable":false,"stateMutability":"view","type":"function"},'
                       '{"anonymous":false,"inputs":[{"indexed":true,"name":"_from","type":"address"},'
                       '{"indexed":true,"name":"_to","type":"address"},{"indexed":false,"name":"_value",'
                       '"type":"uint256"}],"name":"Transfer","type":"event"},{"anonymous":false,"inputs":[{'
                       '"indexed":true,"name":"_owner","type":"address"},{"indexed":true,"name":"_spender",'
                       '"type":"address"},{"indexed":false,"name":"_value","type":"uint256"}],"name":"Approval",'
                       '"type":"event"}]')
BEP_ABI = json.loads("""[ { "anonymous": false, "inputs": [ { "indexed": true, "internalType": "address", 
"name": "owner", "type": "address" }, { "indexed": true, "internalType": "address", "name": "spender", 
"type": "address" }, { "indexed": false, "internalType": "uint256", "name": "value", "type": "uint256" } ], 
"name": "Approval", "type": "event" }, { "anonymous": false, "inputs": [ { "indexed": true, "internalType": 
"address", "name": "from", "type": "address" }, { "indexed": true, "internalType": "address", "name": "to", 
"type": "address" }, { "indexed": false, "internalType": "uint256", "name": "value", "type": "uint256" } ], 
"name": "Transfer", "type": "event" }, { "constant": true, "inputs": [ { "internalType": "address", "name": "_owner", 
"type": "address" }, { "internalType": "address", "name": "spender", "type": "address" } ], "name": "allowance", 
"outputs": [ { "internalType": "uint256", "name": "", "type": "uint256" } ], "payable": false, "stateMutability": 
"view", "type": "function" }, { "constant": false, "inputs": [ { "internalType": "address", "name": "spender", 
"type": "address" }, { "internalType": "uint256", "name": "amount", "type": "uint256" } ], "name": "approve", 
"outputs": [ { "internalType": "bool", "name": "", "type": "bool" } ], "payable": false, "stateMutability": 
"nonpayable", "type": "function" }, { "constant": true, "inputs": [ { "internalType": "address", "name": "account", 
"type": "address" } ], "name": "balanceOf", "outputs": [ { "internalType": "uint256", "name": "", "type": "uint256" } 
], "payable": false, "stateMutability": "view", "type": "function" }, { "constant": true, "inputs": [], 
"name": "decimals", "outputs": [ { "internalType": "uint256", "name": "", "type": "uint256" } ], "payable": false, 
"stateMutability": "view", "type": "function" }, { "constant": true, "inputs": [], "name": "getOwner", "outputs": [ { 
"internalType": "address", "name": "", "type": "address" } ], "payable": false, "stateMutability": "view", 
"type": "function" }, { "constant": true, "inputs": [], "name": "name", "outputs": [ { "internalType": "string", 
"name": "", "type": "string" } ], "payable": false, "stateMutability": "view", "type": "function" }, { "constant": 
true, "inputs": [], "name": "symbol", "outputs": [ { "internalType": "string", "name": "", "type": "string" } ], 
"payable": false, "stateMutability": "view", "type": "function" }, { "constant": true, "inputs": [], 
"name": "totalSupply", "outputs": [ { "internalType": "uint256", "name": "", "type": "uint256" } ], "payable": false, 
"stateMutability": "view", "type": "function" }, { "constant": false, "inputs": [ { "internalType": "address", 
"name": "recipient", "type": "address" }, { "internalType": "uint256", "name": "amount", "type": "uint256" } ], 
"name": "transfer", "outputs": [ { "internalType": "bool", "name": "", "type": "bool" } ], "payable": false, 
"stateMutability": "nonpayable", "type": "function" }, { "constant": false, "inputs": [ { "internalType": "address", 
"name": "sender", "type": "address" }, { "internalType": "address", "name": "recipient", "type": "address" }, 
{ "internalType": "uint256", "name": "amount", "type": "uint256" } ], "name": "transferFrom", "outputs": [ { 
"internalType": "bool", "name": "", "type": "bool" } ], "payable": false, "stateMutability": "nonpayable", 
"type": "function" } ]""")


import colored


class PrettyText:
    def __init__(self, verbosity):
        self.verbosity = verbosity

    def debug(self, data):
        if self.verbosity >= 2:
            print(colored.fore.WHITE+ '[' + colored.fore.GREY_3 + '%' + colored.fore.WHITE + '] ' + colored.style.RESET +
                  str(data))

    def status(self, data):
        print(
            colored.fore.DARK_OLIVE_GREEN_1A + '[' + colored.fore.LIGHT_STEEL_BLUE_1 + 'i' +
            colored.fore.DARK_OLIVE_GREEN_1A + '] ' + colored.style.RESET + str(data))

    def verbose(self, data):
        if self.verbosity >= 1:
            print(colored.fore.LIGHT_YELLOW + '[' + colored.fore.MAGENTA + '*' + colored.fore.LIGHT_YELLOW + ']' +
                  colored.style.RESET + str(data))

    def normal(self, data):
        print(colored.fore.LIGHT_BLUE + colored.style.BOLD + '[' + colored.fore.RED + '+'
              + colored.fore.LIGHT_BLUE + '] ' + colored.style.RESET + str(data))

    def error(self, data):
        print(
            colored.fore.RED_1 + colored.style.BOLD + '[' + colored.fore.WHITE + '!' \
            + colored.fore.RED_1 + '] '+ colored.style.RESET + str(data))

    def good(self, data):
        print(
            colored.fore.LIGHT_GREEN + colored.style.BOLD + '[' + colored.fore.MAGENTA + '~' \
            + colored.fore.LIGHT_GREEN +  '] ' + colored.style.RESET + str(data))

    def warning(self, data):
        print(colored.fore.VIOLET + colored.style.BOLD + '[' + colored.fore.VIOLET + \
              '*' + colored.fore.VIOLET + '] ' + colored.style.RESET + str(data))


#!/usr/bin/env python3
import argparse
import json
import os
import pprint
import time
import uuid

import dotenv
import requests
import web3
from eth_typing import HexStr
from web3.contract import Contract
from web3.exceptions import TransactionNotFound
from web3.middleware import geth_poa_middleware
from web3.types import RPCResponse

import lib.style
from lib import abi_lib, inputs
from lib.w3_validation import validate_addr, valid_token
from lib.wallet_manager import WalletManager


class SecureWeb3:
    def __init__(self, wallet_file: str = None, network: str = 'ethereum'):
        dotenv.load_dotenv()
        self.endpoint = None
        self.token_abi = None
        self.account = None
        self.wallet = None
        self.flashbots_endpoint = None
        self.tokens = []
        self.printer = lib.style.PrettyText(0)
        self.wallet_file = wallet_file
        self.network = network
        self.w3 = self.setup_w3()
        self.eth_price = 0
        self._session = requests.Session()

    def setup_w3(self) -> web3.Web3:
        self.flashbots_endpoint = os.environ.get(f'flashbots_{self.network}_endpoint')
        w3_endpoint = os.environ.get(f'{self.network}_http_endpoint')
        if self.flashbots_endpoint:
            self.printer.normal('Flashbots RPC available')
            w3_endpoint = self.flashbots_endpoint
        self.w3 = web3.Web3(web3.HTTPProvider(w3_endpoint))
        if self.w3.isConnected:
            self.printer.good(f"Connected to chain: {self.w3.eth.chain_id}")
        else:
            self.printer.error(f'Web3 could connect to remote endpoint: {w3_endpoint}')
        if self.network == 'ethereum':
            self.token_abi = lib.abi_lib.EIP20_ABI
        elif self.network == 'polygon':
            self.token_abi = lib.abi_lib.EIP20_ABI
            self.w3.middleware_onion.inject(geth_poa_middleware, layer=0)
        elif self.network == 'bsc':
            self.token_abi = lib.abi_lib.BEP_ABI
            # self.w3.middleware_onion.inject(geth_poa_middleware, layer=0)
            self.endpoint = 'https://bsc.api.0x.org/'
            self.printer.warning('Connected to BSC, which has not been tested very well yet.')
        elif self.network == 'aurora':
            self.token_abi = lib.abi_lib.EIP20_ABI

        self.printer.good(f'Web3 connected to chain: {self.w3.eth.chain_id}')
        return self.w3

    def load_wallet(self) -> bool:
        if not self.wallet_file:
            self.wallet_file = os.environ.get('default_wallet_location')
        self.wallet = WalletManager(self.wallet_file)
        # w.setup_wizard()
        denc, conf = self.wallet.decrypt_load_wallet()
        # print(conf)
        self.account = web3.Account.from_key(denc)
        tokens = conf.get('tokens')
        if tokens is None:
            self.tokens = []
        else:
            self.tokens = tokens
        if self.account:
            # overwrite in memory
            # w = uuid.uuid4().hex
            denc = uuid.uuid4().hex
            # del w
            del denc
            return True
        return False

    def configure_wallet(self) -> None:
        w = WalletManager(self.wallet_file)
        w.setup_wizard()
        w = uuid.uuid4().hex
        del w

    @property
    def web3(self) -> web3.Web3:
        return self.w3

    def query_eth_price(self) -> None:
        if self.w3.eth.chain_id == 0 or self.w3.eth.chain_id == 5:
            s, r = self._session.get(
                url=f'https://api.etherscan.io/api?module=stats&action=ethprice&apikey='
                    f'{os.environ.get("etherscan_api_key")}')
            eth_price = json.loads(json.dumps(r))
            self.eth_price = eth_price.get('result').get('ethusd')
        elif self.w3.eth.chain_id == 56:
            s, r = self._session.get(
                f'https://api.bscscan.com/api?module=stats&action=bnbprice&apikey='
                f'{os.environ.get("bscan_api_key")}')
            eth_price = json.loads(json.dumps(r))
            self.eth_price = eth_price.get('result').get('ethusd')
        elif self.w3.eth.chain_id == 137:
            r = requests.get(
                'https://api.polygonscan.com/' + 'api' +
                f'?module=stats&action=maticprice&apikey={os.environ.get("polygonscan_api_key")}')
            r = r.json()
            eth_price = json.loads(json.dumps(r))
            self.eth_price = eth_price.get('result').get('maticusd')

    def query_gas_api(self) -> (int, int):
        """curl -H 'Authorization: f5534c3e-c1e0-477d-9978-412b9a1276a6'
        'https://api.blocknative.com/gasprices/blockprices?chainid=137'"""
        if self.w3.eth.chain_id == 5:
            chainId = 1
        else:
            chainId = self.w3.eth.chain_id
        ret = self._session.get('https://api.blocknative.com/gasprices/blockprices',
                                params={'chainid': chainId},
                                headers={'Authorization': os.environ.get('blocknative_api_key')})
        if ret.status_code == 200:
            ret = ret.json()
        max_priority_fee_per_gas = ret.get('blockPrices')[0].get('estimatedPrices')[0].get(
            'maxPriorityFeePerGas')
        max_fee_per_gas = ret.get('blockPrices')[0].get('estimatedPrices')[0].get('maxFeePerGas')
        return max_priority_fee_per_gas, max_fee_per_gas

    def switch_network(self, network_name: str, poa: bool = False) -> None:
        """
        Switch network of w3 connections
        :param network_name:
        :param poa: load point of authority middleware for networks like polygon
        """
        endpoint = os.environ.get(f'infura_{network_name}_endpoint')
        if not endpoint:
            self.printer.error('Could not find network, is it configured?')
            return
        w3 = web3.Web3(web3.HTTPProvider(endpoint))
        if network_name == 'polygon':
            poa = True
        if poa:
            w3.middleware_onion.inject(geth_poa_middleware)
        if w3.isConnected():
            self.printer.good(f'Successfully switched to {network_name}')
        else:
            self.printer.error(f'Web3 could not connect to endpoint at {endpoint}')


class EtherShellWallet:
    def __init__(self, sw3: SecureWeb3):
        self.sw3 = sw3
        self._balance = 0

    def build_private_tx(self, tx: dict):
        block = self.sw3.w3.eth.get_block_number() + 1
        _tx = {
            "jsonrpc": "2.0",
            "id": 1,
            "method": "eth_sendPrivateTransaction",
            "params": [{"tx": tx, "maxBlockNumber": block}]}
        return _tx

    def broadcast_raw_tx(self, tx: dict, private: bool = False) -> (HexStr, RPCResponse):
        """
        Send a raw transaction to the rpc for broadcast
        :param tx: Signed Transaction
        :param private: send via private transaction method
        :return: transaction hash
        """
        tx['nonce'] = self.sw3.w3.eth.get_transaction_count(sw3.account.address)
        self.sw3.printer.normal(f'Loaded TX:\n{tx}')
        sign = input('Broadcast? y/n >> ')
        if sign.lower() in ['yes', 'y']:
            signed_tx = self.sw3.account.sign_transaction(tx)
            if private:

                payload = self.build_private_tx(tx)

                headers = {
                    "accept": "application/json",
                    "content-type": "application/json"
                }

                response = requests.post(self.sw3.flashbots_endpoint, json=payload, headers=headers)
                resp = response.json()
                print(response.status_code, resp)
                return resp.get('result').get('bundleHash')
            return web3.Web3.toHex(self.sw3.w3.eth.send_raw_transaction(signed_tx.rawTransaction))

    def _token(self, token_address: str) -> Contract:
        """
        Create and return a Contract object with the ERC/BEP/20 abi
        :param token_address: contract address string
        :return: Contract object
        """
        if self.sw3.w3.eth.chain_id == 56:
            token = self.sw3.w3.eth.contract(self.sw3.w3.toChecksumAddress(token_address), abi=lib.abi_lib.BEP_ABI)
        else:
            token = self.sw3.w3.eth.contract(self.sw3.w3.toChecksumAddress(token_address), abi=lib.abi_lib.EIP20_ABI)
        return token

    def eth_balance(self, raw: bool = False) -> (int, float):
        """

        :param raw: if true return raw integer
        :return:
        """
        self._balance = sw3.w3.eth.getBalance(sw3.account.address)
        if raw:
            return self._balance
        return self._balance / (10 ** 18)

    def token_convert(self, token_address: str, qty: int) -> float:
        for token in self.sw3.tokens:
            if self.sw3.w3.toChecksumAddress(token_address) == self.sw3.w3.toChecksumAddress(token.get("address")):
                dec = token.get('decimals')
                return int(qty / (10 ** dec))

    def token_balance(self, token_address: str, raw: bool = False) -> (int, float):
        token = self._token(token_address)
        token_balance = token.functions.balanceOf(self.sw3.account.address).call()
        if raw:
            return token_balance
        for tok in self.sw3.tokens:
            if self.sw3.w3.toChecksumAddress(tok.get('address')) == self.sw3.w3.toChecksumAddress(token_address):
                dec = int(tok.get('decimals'))
                return token_balance / (10 ** dec)

    def poll_receipt(self, tx_hash: hex):
        poll = 0
        while True:
            if poll > 100:
                break
            poll += 1
            self.sw3.printer.normal(f'Polling for receipt: {poll}/100 ... ')
            try:
                receipt = self.sw3.w3.eth.get_transaction_receipt(tx_hash)
            except web3.exceptions.TransactionNotFound:
                # await asyncio.sleep(1)
                time.sleep(1)
            else:
                receipt = receipt.__dict__
                return receipt
        self.sw3.printer.error('Timed out, transaction may be underpriced!')

    def send_erc20_token(self, raw_amount: int, destination: str, token_address: str, legacy: bool = False,
                         private: bool = False) -> HexStr:
        token = self.sw3.w3.eth.contract(self.sw3.w3.toChecksumAddress(token_address), abi=lib.abi_lib.EIP20_ABI)
        mpfpg, mfpg = self.sw3.query_gas_api()
        if self.sw3.w3.eth.chain_id in [0, 1, 5, 137]:
            tx = {
                'maxPriorityFeePerGas': self.sw3.w3.toWei(mpfpg, 'gwei'),
                'maxFeePerGas': self.sw3.w3.toWei(mfpg, 'gwei'),
                "to": self.sw3.w3.toChecksumAddress(destination),
                "value": "0x0",
                "data": token.encodeABI('transfer', args=(self.sw3.w3.toChecksumAddress(destination), raw_amount)),
                "nonce": self.sw3.w3.eth.get_transaction_count(self.sw3.account.address),
                "chainId": self.sw3.w3.eth.chain_id
            }
            if legacy:
                tx.pop('maxFeePerGas')
                tx.pop('maxPriorityFeePerGas')
                tx.pop('type')
                tx.update({'gasPrice': self.sw3.w3.toWei(int(self.sw3.w3.eth.gas_price * 1.1), 'gwei')})

            gas = self.sw3.w3.eth.estimate_gas(tx)
            tx.update({'gas': gas})
            return self.broadcast_raw_tx(tx, private)

    def send_eth(self, raw_amount: int, destination: str, legacy: bool = False, private: bool = False) -> HexStr:
        mpfpg, mfpg = self.sw3.query_gas_api()

        tx = {
            'nonce': self.sw3.w3.eth.get_transaction_count(self.sw3.account.address),
            'to': self.sw3.w3.toChecksumAddress(destination),
            'value': raw_amount,
            'maxFeePerGas': self.sw3.w3.toWei(mfpg, 'gwei'),
            'maxPriorityFeePerGas': self.sw3.w3.toWei(mpfpg, 'gwei'),
            'type': 2,
            'chainId': hex(self.sw3.w3.eth.chain_id)
        }
        if legacy:
            tx.pop('maxFeePerGas')
            tx.pop('maxPriorityFeePerGas')
            tx.pop('type')
            tx.update({'gasPrice': self.sw3.w3.toWei(int(self.sw3.w3.eth.gas_price * 1.1), 'gwei')})
        gas = self.sw3.w3.eth.estimate_gas(tx)
        tx.update({'gas': gas})
        return self.broadcast_raw_tx(tx, private)

    def import_token(self, token_address: str) -> bool:
        if not validate_addr(token_address):
            return False
        if not valid_token(self._token(token_address)):
            return False
        token = self._token(token_address)
        decimals = token.functions.decimals().call()
        symbol = token.functions.symbol().call()
        balance = token.functions.balanceOf(self.sw3.account.address).call()
        token_dict = {'address': self.sw3.w3.toChecksumAddress(token_address),
                      'network': self.sw3.w3.eth.chain_id,
                      'decimals': decimals,
                      'symbol': symbol,
                      'balance': balance}
        self.sw3.tokens.append(token_dict)
        self.sw3.wallet.wallet.update_wallet('tokens', self.sw3.tokens)
        return True

    def interactive(self, action: str = 'send', _type: str = 'eth', legacy: bool = False, private: bool = False):
        # print(action, _type, legacy)
        if action == 'send':
            balance = self.eth_balance()
            self.sw3.printer.normal(f'Launching interactive wallet shell.\n')
            self.sw3.printer.good(f'Account Balance: {balance} ')
            destination = lib.inputs.get_dest_addr()
            if self.sw3.w3.eth.getCode(destination):
                self.sw3.printer.warning('Warning! This is a contract address!')
            if _type == 'eth':
                while True:
                    amount = input("Amount in ether >> ")
                    amount = float(amount)
                    if float(amount):
                        if amount <= balance:
                            break
                        else:
                            self.sw3.printer.error(f'Amount exceeds current wallet balance: {balance}')
                self.sw3.printer.normal(f'Transaction parameters: \nSend {amount} to {destination}')
                if lib.inputs.confirmation():
                    amount = amount * (10 ** 18)
                    txid = self.send_eth(int(amount), destination, legacy, private)
                    self.sw3.printer.good(f'TXID: {txid}')
                    receipt = self.poll_receipt(txid)
                    if receipt:
                        pprint.pprint(receipt)
            if _type == 'erc20':
                self.sw3.printer.normal('Select a token to send: ')
                while True:
                    for x, token in enumerate(self.sw3.tokens):
                        addr = token.get('address')
                        token_balance = self.token_convert(addr, token.get('balance'))
                        addr = addr[:4] + '..' + addr[-4:]
                        print(f'[{x}], {token.get("symbol")}@{addr}, Balance: {token_balance} (local)')
                    selection = input('Selection # >> ')
                    if int(selection) <= int(len(self.sw3.tokens) - 1):
                        token = self.sw3.tokens[int(selection)]
                        symbol = token.get('symbol')
                        token_address = token.get('address')
                        token_short_addr = token_address[:4] + '..' + token_address[-4:]
                        token_balance = self.token_balance(token_address)
                        self.sw3.printer.normal(f'Balance: {token_balance} (live)')
                        break
                while True:
                    amount = input(f"Amount in {symbol} >> ")
                    amount = float(amount)
                    if float(amount):
                        if amount <= token_balance:
                            break
                        else:
                            self.sw3.printer.error(f'Amount exceeds current wallet balance: {token_balance}')
                qty = int(amount * (10 ** int(token.get('decimals'))))
                if balance <= 0:
                    self.sw3.printer.error('Not enough ETH to pay for gas!')
                    return False
                self.sw3.printer.normal(f'Transaction parameters: \nSend {amount} of {symbol} @{token_short_addr} '
                                        f'to {destination}')
                if lib.inputs.confirmation():
                    txid = self.send_erc20_token(qty, destination, token.get('address'), legacy, private)
                    if txid:
                        receipt = self.poll_receipt(txid)
                        pprint.pprint(receipt)


if __name__ == '__main__':
    args = argparse.ArgumentParser('Secure Web3 Cli')
    general_opts = args.add_argument_group(description='Wallet managment options.')
    general_opts.add_argument('-i', '--init', dest='init_wallet', type=str, nargs=1,
                              default=None, help='Initialize this new wallet')
    general_opts.add_argument('-o', '--open', dest='open_wallet', type=str, default=None,
                              nargs='?', help='Unlock a wallet, use default wallet if not specified.')
    general_opts.add_argument('-l', '--lock', dest='lock_wallet', action='store_true',
                              help='Lock all open wallets.')
    general_opts.add_argument('-n', '--network', type=str, default='ethereum',
                              choices=['ethereum', 'polygon', 'bsc', 'aurora', 'goerli'])
    config_options = args.add_argument_group(description='Wallet configuration options and functions')
    config_options.add_argument('-it', '--import-token', dest='import_token', type=str,
                                default=None, help='Add this token to the specified wallet.')
    read_opts = args.add_argument_group(description='EVM State-Reading functions.')
    read_opts.add_argument('-b', '--balance', action='store_true', help='Get wallet balance info.')
    tx_options = args.add_argument_group(description='EVM State-Writing related options.')
    tx_options.add_argument('-r', '--raw', dest='broadcast_raw', type=str, help='Load a json tx from '
                                                                                'this file to sign and broadcast.')
    tx_options.add_argument('-s', '--send', choices=['eth', 'erc20'], default=None,
                            help='Open interactive shell to send ethereum.')
    tx_options.add_argument('-p', '--private', dest='use_private_tx', action='store_true',
                            help='Submit transaction private via flashbots (if available).')

    tx_options.add_argument('-L', '--legacy', action='store_true', default=False, help='Use legacy gas protocol.')

    args = args.parse_args()
    dotenv.load_dotenv()

    if args.init_wallet:
        print(f'[+] Configuring new wallet "{args.init_wallet[0]}" ... ')
        manager = WalletManager(args.init_wallet[0])
        manager.setup_wizard()
        exit(0)
    if not args.open_wallet:
        wallet_file = os.environ.get('default_wallet_location')
    else:
        wallet_file = args.open_wallet
    sw3 = SecureWeb3(wallet_file, args.network)
    sw3.printer.normal(f'Opening wallet "{wallet_file}" ... ')
    sw3.load_wallet()
    wallet = EtherShellWallet(sw3)
    if args.import_token:
        wallet.import_token(args.import_token)
    if args.balance:
        ether_balance = wallet.eth_balance()
        sw3.printer.normal(f'Ethereum Balance: {ether_balance}')
    if args.broadcast_raw:
        with open(args.broadcast_raw, 'r') as f:
            tx = json.load(fp=f)
        txid = wallet.broadcast_raw_tx(tx, private=args.private)
        sw3.printer.normal(f'TXID: {txid}')
        wallet.poll_receipt(txid)
    if args.send:
        wallet.interactive(action='send', _type=args.send, legacy=args.legacy, private=args.use_private_tx)



from getpass import getpass
from typing import Tuple
import os

import hmac
import json

import base64
import hashlib
import uuid

import dotenv
import web3
from Crypto import Random
from Crypto.Cipher import AES


class JsonWallet:
    def __init__(self, wallet_file, wallet_file_env=None):
        dotenv.load_dotenv()
        if wallet_file_env:
            self.wallet_file = os.environ.get(wallet_file)
        else:
            self.wallet_file = wallet_file
        self.jw = {'enc_password': None,
                   'salt': None,
                   'enc_wallet': None,
                   'public_address': None,
                   'tokens': [],
                   'comment': None}

    def setup(self, encrypted_password: bytes = None,
              salt: bytes = None,
              encrypted_key: bytes = None,
              public_key: str = None,
              comment: str = None):
        # encoded_password =
        self.jw.update({'enc_password': base64.b64encode(encrypted_password).decode()})
        self.jw.update({'salt': base64.b64encode(salt).decode()})
        self.jw.update({'enc_wallet': base64.b64encode(encrypted_key).decode()})
        self.jw.update({'public_address': public_key})
        self.jw.update({'native_balances': {}})
        self.jw.update({'tokens': []})
        self.jw.update({'comment': comment})

    def save_wallet(self):
        with open(self.wallet_file, 'w') as f:
            json.dump(obj=self.jw, fp=f)

    def update_wallet(self, field, value):
        #
        if self.jw.get(field):
            setattr(self.jw, field, value)
        else:
            self.jw.update({field: value})
        self.save_wallet()

    def load_wallet(self):
        with open(self.wallet_file, 'r') as f:
            self.jw = json.load(fp=f)
            # print(self.jw)
            # Fix older wallets missing fields
            if type(self.jw.get('tokens')) is None:
                self.jw.update({'tokens': []})
                self.save_wallet()
            if type(self.jw.get('comment')) is None:
                self.jw.update({'comment': ''})
                self.save_wallet()
            if type(self.jw.get('native_balances')) is None:
                self.jw.update({'native_balances': []})
                self.save_wallet()

            return base64.b64decode(self.jw.get('enc_password')), \
                   base64.b64decode(self.jw.get('salt')), \
                   base64.b64decode(self.jw.get('enc_wallet')), \
                   self.jw.get('public_address'), \
                   self.jw.get('native_balances'), \
                   self.jw.get('tokens'), \
                   self.jw.get('comment')


class AESCipher(object):

    def __init__(self, key):
        self.bs = AES.block_size
        self.key = hashlib.sha256(key.encode()).digest()

    def encrypt(self, raw):
        raw = self._pad(raw)
        iv = Random.new().read(AES.block_size)
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return base64.b64encode(iv + cipher.encrypt(raw.encode()))

    def decrypt(self, enc):
        enc = base64.b64decode(enc)
        iv = enc[:AES.block_size]
        cipher = AES.new(self.key, AES.MODE_CBC, iv)
        return self._unpad(cipher.decrypt(enc[AES.block_size:])).decode('utf-8')

    def check_encryption(self, enc, plaintext):
        dec = self.decrypt(enc)
        assert dec == plaintext

    def _pad(self, s):
        return s + (self.bs - len(s) % self.bs) * chr(self.bs - len(s) % self.bs)

    @staticmethod
    def _unpad(s):
        return s[:-ord(s[len(s) - 1:])]


class HasherLib:

    def hash_str(self, password: str) -> Tuple[bytes, bytes]:
        """
        Hash the provided password with a randomly-generated salt and return the
        salt and hash to store in the database.
        """
        salt = os.urandom(16)
        pw_hash = hashlib.pbkdf2_hmac('sha256', password.encode(), salt, 100000)
        return salt, pw_hash

    def check_hash_str(self, salt: bytes, pw_hash: bytes, password: str) -> bool:
        """
        Given a previously-stored salt and hash, and a password provided by a user
        trying to log in, check whether the password is correct.
        """
        return hmac.compare_digest(
            pw_hash,
            hashlib.pbkdf2_hmac('sha256', password.encode(), salt, 100000)
        )


class WalletManager:
    def __init__(self, wallet_file):
        self.wallet_file = wallet_file
        self.wallet = JsonWallet(wallet_file)
        self.aes = AESCipher
        self.hasher = HasherLib()

    def privkey_to_address(self, key):
        acct = web3.eth.Account.from_key(str(key))
        return acct.address

    def setup_wizard(self):
        password = ''
        confirm_password = ''
        # Example usage:
        for x in range(3):
            password: str = getpass('Enter password  >> ')
            confirm_password = str(getpass('Confirm password >>: '))
            if password == confirm_password:
                break
            else:
                print('Passwords do not match')
        if not password:
            return False

        for x in range(3):
            s, p = self.hasher.hash_str(password=password)
            assert self.hasher.check_hash_str(s, p, password)
            assert not self.hasher.check_hash_str(s, p, str(uuid.uuid4().hex))
            private_key = getpass('Enter privkey >> ')
            confirm_key = getpass('Confirm key >> ')
            if private_key == confirm_key:
                self.aes = AESCipher(password)
                enc_key = self.aes.encrypt(private_key)
                try:
                    self.aes.check_encryption(enc_key, private_key)
                except AssertionError:
                    print('[!] Error decrypting, try again.')
                else:
                    print('[+] The wizard successfully configured your wallet.')
                    comment = str(input('Memo >>'))
                    if not comment:
                        comment = ''
                    if password == confirm_password:
                        print('Success ... encrypting')

                        self.wallet.setup(encrypted_password=p, salt=s, encrypted_key=enc_key,
                                          public_key=self.privkey_to_address(private_key),
                                          comment=comment)

                        del password, confirm_password, private_key, confirm_key, comment
                        self.wallet.save_wallet()

                        break
                    else:
                        print('Keys did not match ...')

    def _decrypt_load_wallet(self):
        """
        Possibly more secure method?
        :return:
        """
        pw = getpass('Unlock wallet >>')
        salt, password, enc_key, address, native_balances, tokens, comment = self.wallet.load_wallet()

        ret_dict = {"tokens": tokens, "comment": comment, "balances": native_balances}
        if self.hasher.check_hash_str(password, salt, pw):
            self.aes = AESCipher(pw)
            return self.aes.decrypt(enc=enc_key), ret_dict
        else:
            return False

    def decrypt_load_wallet(self):
        """
        User friendly method
        :return:
        """
        ret = {}
        denc = False
        for x in range(3):
            try:
                denc, ret = self._decrypt_load_wallet()
            except KeyboardInterrupt:
                print('[-] Caught Signal, exit gracefully .. ')
            else:
                if ret:
                    print('[~] Successfully unlocked wallet .. ')
                    break
                else:
                    print('[!] Incorrect password')
            if x + 1 == 3:
                print('Max tries exceeded.')
        return denc, ret


if __name__ == '__main__':
    w = WalletManager('.wallet')
    w.setup_wizard()



from setuptools import setup

setup(
    name='secure_web3',
    version='1.15',
    packages=['secure_web3'],
    install_requires=['web3', 'pycryptodome', 'dotenv', 'eth-utils'],
    package_dir={"": "src"},
    include_package_data=True,
    data_files=[('', ['src/secure_web3/keys/default_wallet.json', 'src/secure_web3/data/networks.json'])],
    url='https://github.com/darkerego/secure_web3',
    license='MIT',
    author='darkerego',
    author_email='chevisyoung@gmail.com',
    description='Secure Web3 Wallet and DevEnv'
)
